#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdint.h>

#include <openssl/hmac.h>

static int hmac_sha256(uint8_t *message, int message_len)
{
    uint8_t key[] = "LUM{0qvYH:QI?K6}";

    uint8_t digest[] = {
          0x50, 0x48, 0x61, 0x08, 0x9d, 0x10, 0xa8, 0xd5, 0x90, 0x0c, 0xdf, 0x3b,
            0xad, 0x96, 0x20, 0x04, 0x28, 0x2e, 0x73, 0xc2, 0x0d, 0x2d, 0x5a, 0xb9,
              0x5b, 0x67, 0xae, 0x6b, 0x33, 0x56, 0x74, 0x8b
    };

    uint8_t out[EVP_MAX_MD_SIZE];
    uint32_t outlen;

    if (!HMAC(EVP_sha256(),key,sizeof(key)-1,message,message_len,out,&outlen)) {
        assert(0);
    }

    if (memcmp(digest, out, outlen) == 0) {
        for (int i=0; i<message_len; i++) {
            printf("%hhx", message[i]);
        }
        printf("\n");
        return 0;
    }
    return 1;
}

uint32_t A = 0XADAAA9AA;
uint32_t B = 0X52555655;

void get_binary(uint32_t n, uint8_t l[]) {
    memset(l, 0xff, 32);
    for (int i=0; i<32; i++) {
        if (((n >> i) & 0x1) == 1) {
            l[i] = i & 0xff;
        }
    }
}

void get_first(uint8_t l[], uint8_t right_shift) {
    uint8_t copy[32];
    memcpy(copy, l, 32);
    for (int i=0; i<32; i++) {
        if (i < (32 - right_shift))
            l[i] = copy[i + right_shift];
        else
            l[i] = copy[i + right_shift - 32];
    }
}

void get_second(uint8_t l[], uint8_t rshift2) {
    uint8_t copy[32];
    memcpy(copy, l, 32);
    memset(l, 0xff, 32);
    for (int i=0; i<32-rshift2; i++) {
        l[i] = copy[i + rshift2];
    }
}

void do_transform(uint32_t n, uint8_t l[], uint8_t right_shift, uint8_t rshift2) {
    get_binary(n, l);
    get_first(l, right_shift);
    get_second(l, rshift2);
}

void get_value_bits(uint8_t v, uint8_t l[]) {
    for (int i=0; i<8; i++) {
        l[i] = (v >> i) & 0x1;
    }
}

void step(uint8_t value, uint8_t key_idx_A, uint8_t key_idx_B, uint8_t right_shift, uint8_t rshift2, uint32_t offset, uint8_t keybits[]) {
    uint8_t lA[32];
    uint8_t lB[32];
    do_transform(A, lA, right_shift, rshift2);
    do_transform(B, lB, right_shift, rshift2);

    uint8_t v = (value - offset) & 0xff;
    uint8_t vlist[8];
    get_value_bits(v, vlist);

    for (int i=0; i<8; i++) {
        if (vlist[i] == 1) {
            if (lA[i] != 0xff)
                keybits[key_idx_A * 32 + lA[i]] = 1;
            if (lB[i] != 0xff)
                keybits[key_idx_B * 32 + lB[i]] = 1;
        }
    }
}

uint8_t a5 = 251;
uint8_t s2 = 238;
uint8_t s1 = 77;
uint8_t t2 = 75;
uint8_t t0 = 126;
uint8_t s4 = 59;
uint8_t t6 = 71;
uint8_t t5 = 134;
uint8_t s7 = 22;
uint8_t t3 = 43;
uint8_t a7 = 158;
uint8_t a6 = 44;
uint8_t a0 = 238;
uint8_t a1 = 221;

// return [a5, s2, s1, t2, t0, s4, t6, t5, s7, t3, a7, a6, a0, a1, a2, t1]
// intermediate = [251, 238, 77, 75, 126, 59, 71, 134, 22, 43, 158, 44, 238, 221]

void get_hex(uint8_t hex[], uint8_t keybits[]) {
    for (int i=0; i<16; i++) {
        hex[i] = 0;
        for (int j=0; j<8; j++) {
            hex[i] |= (keybits[8 * i + j] << j);
        }
    }
}



int main(void) {
    uint8_t keybits[128];
    uint8_t hex[16];
    for (int a2=0; a2<256; a2++) {
        for (int t1=0; t1<256; t1++) {
            memset(keybits, 0, 128);
            memset(hex, 0, 16);
            step(t1, 3, 1, 23, 24,       65, keybits);
            step(a2, 3, 1, 23, 16, t1 +  72, keybits);
            step(a1, 3, 1, 23,  8, a2 +  79, keybits);
            step(a0, 3, 1, 23,  0, a1 +  86, keybits);
            step(a6, 2, 0, 19, 24, a0 +  93, keybits);
            step(a7, 2, 0, 19, 16, a6 + 100, keybits);
            step(t3, 2, 0, 19,  8, a7 + 107, keybits);
            step(s7, 2, 0, 19,  0, t3 + 114, keybits);
            step(t5, 1, 3, 17, 24, s7 + 121, keybits);
            step(t6, 1, 3, 17, 16, t5 + 128, keybits);
            step(s4, 1, 3, 17,  8, t6 + 135, keybits);
            step(t0, 1, 3, 17,  0, s4 + 142, keybits);
            step(t2, 0, 2, 13, 24, t0 + 149, keybits);
            step(s1, 0, 2, 13, 16, t2 + 156, keybits);
            step(s2, 0, 2, 13,  8, s1 + 163, keybits);
            step(a5, 0, 2, 13,  0, s2 + 170, keybits);
            get_hex(hex, keybits);
            if (hmac_sha256(hex, 16) == 0) {
                exit(0);
            }
        }
    }
    return 0;
}





