import struct

import instructions_list
import instructions_list_ida

START_OFFSET = 0x05571C00

PC = START_OFFSET

def read_unsigned_byte():
    global PC
    b = struct.unpack('B', INSTRUCTIONS[PC-START_OFFSET])[0]
    PC += 1
    return b

def read_four_unsigned_bytes():
    global PC
    off = PC - START_OFFSET
    i = struct.unpack('<I', INSTRUCTIONS[off:off+4])[0]
    PC += 4
    return i

# http://stackoverflow.com/questions/8461798/how-can-i-struct-unpack-many-numbers-at-once
fmt = "%dB" % len(instructions_list_ida.INSTRUCTIONS_LIST_IDA)
INSTRUCTIONS = struct.pack(fmt, *instructions_list_ida.INSTRUCTIONS_LIST_IDA)
#fmt = "%dB" % len(instructions_list.INSTRUCTIONS_LIST)
#INSTRUCTIONS = struct.pack(fmt, *instructions_list.INSTRUCTIONS_LIST)

def do_mov_imm(pc, b):
    i = read_four_unsigned_bytes()
    imm = i & 0x1fffffff
    dst = i >> 0x1d
    print '%#x: mov r%d, %#x' % (pc, dst, imm)

def do_mov_reg(pc, b):
    b2 = read_unsigned_byte()
    dst = (b2 >> 4) & 7
    src = (b2 >> 1) & 7
    print '%#x: mov r%d, r%d' % (pc, dst, src)

def do_mov_reg2(pc, b):
    b2 = read_unsigned_byte()
    dst = (b2 >> 4) & 7
    src = (b2 >> 1) & 7
    print '%#x: mov [r%d], r%d' % (pc, dst, src)

def do_mov_reg3(pc, b):
    b2 = read_unsigned_byte()
    dst = (b2 >> 4) & 7
    src = (b2 >> 1) & 7
    print '%#x: mov r%d, [r%d]' % (pc, dst, src)

MOV_OPCODES = {
        0 : do_mov_imm,
        1 : do_mov_reg,
        2 : do_mov_reg2,
        3 : do_mov_reg3,
        }

def do_mov(pc, b):
    MOV_OPCODES[b & 3](pc, b)

def do_add_imm(pc, b):
    i = read_four_unsigned_bytes()
    imm = i & 0x1fffffff
    dst = i >> 0x1d
    print '%#x: add r%d, %#x' % (pc, dst, imm)

def do_add_reg(pc, b):
    b2 = read_unsigned_byte()
    dst = (b2 >> 4) & 7
    src = (b2 >> 1) & 7
    print '%#x: add r%d, r%d' % (pc, dst, src)

ADD_OPCODES = {
        0 : do_add_imm,
        1 : do_add_reg,
        }

def do_add(pc, b):
    ADD_OPCODES[(b >> 1) & 1](pc, b)

def do_call(pc, b):
    b2 = read_unsigned_byte()
    c = (b & 7) << 8 | b2
    if ((c & 0x400) != 0):
        # negative number
        addr = PC - (c & 0x3ff)
    else:
        addr = PC + c
    print '%#x: call %#x' % (pc, addr)

def do_jmp(pc, b):
    b2 = read_unsigned_byte()
    c = (b & 7) << 8 | b2
    if ((c & 0x400) != 0):
        # negative number
        addr = PC - (c & 0x3ff)
    else:
        addr = PC + c
    print '%#x: jmp %#x' % (pc, addr)

def do_push_imm(pc, b):
    imm = read_four_unsigned_bytes()
    print '%#x: push %#x' % (pc, imm)

def do_push_reg(pc, b):
    b2 = read_unsigned_byte()
    r = (b2 >> 3) & 7
    print '%#x: push r%d' % (pc, r)

PUSH_OPCODES = {
        0 : do_push_imm,
        1 : do_push_reg,
        }

def do_push(pc, b):
    PUSH_OPCODES[(b >> 2) & 1](pc, b)

def do_loc_402F42(pc, b):
    print '%#x: loc_402F42' % pc

def do_loc_4030AF_change_y(pc, b):
    print '%#x: loc_4030AF_change_y' % pc

def do_ret(pc, b):
    print '%#x: ret' % pc

def do_loc_403017_GetPixel(pc, b):
    print '%#x: loc_403017 (GetPixel)' % pc

def do_je_imm(pc, b):
    regs = read_four_unsigned_bytes()
    r0 = regs >> 0x1D
    val = (regs >> 0x13) & 0xff
    c = (regs >> 6) & 0x7ff
    if ((c & 0x400) != 0):
        # negative number
        addr = PC - (c & 0x3ff)
    else:
        addr = PC + c
    print '%#x: cmp r%d %#x, je %#x' % (pc, r0, val, addr)

def do_je_reg(pc, b):
    regs = read_four_unsigned_bytes()
    r0 = regs >> 0x1D
    r1 = (regs >> 0x12) & 7
    c = (regs >> 6) & 0x7ff
    if ((c & 0x400) != 0):
        # negative number
        addr = PC - (c & 0x3ff)
    else:
        addr = PC + c
    print '%#x: cmp r%d r%d, je %#x' % (pc, r0, r1, addr)

JE_OPCODES = {
        0 : do_je_imm,
        1 : do_je_reg,
        }

def do_je(pc, b):
    JE_OPCODES[b & 1](pc, b)

def do_sub_imm(pc, b):
    i = read_four_unsigned_bytes()
    imm = i & 0x1fffffff
    dst = i >> 0x1d
    print '%#x: sub r%d, %#x' % (pc, dst, imm)

def do_sub_reg(pc, b):
    b2 = read_unsigned_byte()
    dst = (b2 >> 4) & 7
    src = (b2 >> 1) & 7
    print '%#x: sub r%d, r%d' % (pc, dst, src)

SUB_OPCODES = {
        0 : do_sub_imm,
        1 : do_sub_reg,
        }

def do_sub(pc, b):
    SUB_OPCODES[(b >> 1) & 1](pc, b)

def do_xor_imm(pc, b):
    i = read_four_unsigned_bytes()
    imm = i & 0x1fffffff
    dst = i >> 0x1d
    print '%#x: xor r%d, %#x' % (pc, dst, imm)

def do_xor_reg(pc, b):
    b2 = read_unsigned_byte()
    dst = (b2 >> 4) & 7
    src = (b2 >> 1) & 7
    print '%#x: xor r%d, r%d' % (pc, dst, src)

XOR_OPCODES = {
        0 : do_xor_imm,
        1 : do_xor_reg,
        }

def do_xor(pc, b):
    XOR_OPCODES[(b >> 1) & 1](pc, b)

def do_loc_403060_call_grosse_fonction(pc, b):
    print '%#x: loc_403060 call sub_402490' % pc

def do_activate_thread_and_sleep_3_seconds(pc, b):
    print '%#x: activate_thread_and_sleep_3_seconds' % pc

def do_andb(pc, b):
    b2 = read_unsigned_byte()
    dst = b & 7
    print '%#x: andb r%d, %#x' % (pc, dst, b2)


def do_shr(pc, b):
    b2 = read_unsigned_byte()
    dst = b & 7
    src = (b2 >> 3) & 7
    print '%#x: shr r%d, r%d' % (pc, dst, src)


def do_shl(pc, b):
    b2 = read_unsigned_byte()
    dst = b & 7
    print '%#x: shl r%d, %#x' % (pc, dst, b2)

OPCODES = {
        0x00 : do_loc_4030AF_change_y,
        0x18 : do_call,
        0x28 : do_loc_403060_call_grosse_fonction,
        0x30 : do_je,
        0x48 : do_activate_thread_and_sleep_3_seconds,
        0x50 : do_andb,
        0x58 : do_ret,
        0x68 : do_xor,
        0x78 : do_loc_402F42,
        0x80 : do_sub,
        0x90 : do_loc_403017_GetPixel,
        0xA0 : do_jmp,
        0xA8 : do_shr,
        0xB0 : do_add,
        0xC8 : do_push,
        0xE0 : do_shl,
        0xE8 : do_mov,
        }

def dump(start_offset, end_offset, execute=True, display=False):
    global PC
    PC = start_offset
    while (PC < end_offset):
        pc = PC
        b = read_unsigned_byte()
        if display:
            print "%#x, %#x" % (b, (b & 0xf8))
        if execute:
            try:
                OPCODES[b & 0xf8](pc, b)
            except:
                pass

dump(START_OFFSET, 0x05571EF9)
dump(0x05571FED, 0x05572000)
