#include <stdint.h>
#include <stdio.h>
#include <inttypes.h>
#include <string.h>
#include <stdbool.h>
#include <stdlib.h>

#if HAVE_BYTESWAP_H
#include <byteswap.h>
#else
#define bswap_16(value) \
    ((((value) & 0xff) << 8) | ((value) >> 8))

#define bswap_32(value) \
    (((uint32_t)bswap_16((uint16_t)((value) & 0xffff)) << 16) | \
     (uint32_t)bswap_16((uint16_t)((value) >> 16)))

#define bswap_64(value) \
    (((uint64_t)bswap_32((uint32_t)((value) & 0xffffffff)) \
      << 32) | \
     (uint64_t)bswap_32((uint32_t)((value) >> 32)))
#endif

uint64_t rol(uint64_t in, uint8_t in_shift)
{
    uint8_t shift = in_shift & 0x3f;
    //printf("%s: shift = %hhu\n", __func__, shift);
    return ((in << shift) | (in >> (64 - shift)));
}

uint64_t ror(uint64_t in, uint8_t in_shift)
{
    uint8_t shift = in_shift & 0x3f;
    //printf("%s: shift = %hhu\n", __func__, shift);
    return ((in >> shift) | (in << (64 - shift)));
}

uint32_t key[] = {
    0X9AACC69B,
    0x89A31D8A,
    0xA5B2CD1C,
    0xB0A5CE54,
};

#define DELTA 0xd2d413b3

#define XOR_CONSTANTE 0xFDE7F446

void decipher(unsigned int num_rounds, uint32_t v[2], uint32_t const key[4]) {
    unsigned int i;
    uint32_t v0=v[0], v1=v[1], delta=DELTA, sum=delta*num_rounds;
    for (i=0; i < num_rounds; i++) {
        v1 -= (((v0 << 4) ^ (v0 >> 5)) + v0) ^ (sum + key[(sum>>11) & 3]);
        sum -= delta;
        v0 -= (((v1 << 4) ^ (v1 >> 5)) + v1) ^ (sum + key[sum & 3]);
    }
    v[0]=v0; v[1]=v1;
}

int revert(uint32_t dword, uint64_t orig[])
{
    uint8_t res[32];
    memset(res, 0, 32);

    int idx = 0;
    for (int j=0; j<4; j++) {
        uint32_t *v = (uint32_t *) &orig[j];
        decipher(64, v, key);
        int i;
        for (i=0; i<2; i++) {
            //printf("v[%d] = %#x\n", i, v[i]);
            uint8_t *chars = (uint8_t *) &v[i];
            for (int k=0; k<4; k++) {
                res[idx] = chars[k];
                idx++;
            }
        }
    }

    for (idx=0; idx<32; idx++) {
        printf("%c", res[idx]);
    }
    printf("\n");

    return 0;
}

unsigned char value_to_xor_with_dword[] =
{
  0x63, 0x93, 0x66, 0xD4, 0x33, 0x9F, 0x03, 0x29, 0xA7, 0x88, 
  0x78, 0xC9, 0x60, 0x1C, 0xFA, 0xCE, 0xC8, 0x90, 0x04, 0x39, 
  0x20, 0x79, 0x82, 0x87, 0x66, 0x1C, 0x84, 0x1A, 0xF9, 0xEE, 
  0x3A, 0x1B
};

void compute_reverse(uint64_t *input, uint32_t dword)
{
    uint64_t rdx = 0x756AF83DE1FFE263;
    uint64_t rcx = 0XFA34AE1002547B89;
    uint64_t orig[4];

    rdx = 0xFFFFFFFFFFFFFFFF - rdx;

    //printf("1.rcx = %" PRIx64 "\n", rcx);
    rcx = bswap_64(rcx);
    //printf("2.rcx = %" PRIx64 "\n", rcx);

    memset(orig, 0, 32);
    orig[0] = (input[0] - rdx) ^ rcx;
    orig[1] = (bswap_64(input[1] + rcx)) ^ rdx;

    rcx = rol(rcx, 0xd);
    orig[2] = bswap_64(rol(input[2], rcx));

    rdx = ror(rdx, 0x15);
    orig[3] = rol(input[3] - rcx, rdx) ^ rcx;

    revert(dword, orig);
}

unsigned char compare_values[] =
{
    0x25, 0x67, 0x81, 0x29, 0x75, 0x6B, 0xE4, 0xD4, 0xE1, 0x7C, 
    0x9F, 0x34, 0x26, 0xE8, 0x1D, 0x33, 0x8E, 0x64, 0xE3, 0xC4, 
    0x66, 0x8D, 0x65, 0x7A, 0x20, 0xE8, 0x63, 0xE7, 0xBF, 0x1A, 
    0xDD, 0xE6
};

int main()
{
    uint32_t dword = XOR_CONSTANTE;
    uint32_t ida_chars[8];
    memset(ida_chars, 0, 32);

    uint32_t *p = (uint32_t *) &value_to_xor_with_dword[0];
    for (int i=0; i<8; i++)
    {
        ida_chars[i] = p[i] ^ dword;
    }

    if (memcmp(ida_chars, compare_values, 32) != 0) {
        exit(-1);
    }

    uint64_t *input = (uint64_t *) &ida_chars[0];
    compute_reverse(input, dword);
    return 0;
}
